package nxt;

import nxt.db.DbUtils;
import nxt.util.Logger;

import java.math.BigInteger;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

final class BlockDb {

	static BlockImpl findBlock(long blockId) {
		try (Connection con = Db.db.getConnection();
				PreparedStatement pstmt = con.prepareStatement("SELECT * FROM block WHERE id = ?")) {
			pstmt.setLong(1, blockId);
			try (ResultSet rs = pstmt.executeQuery()) {
				BlockImpl block = null;
				if (rs.next()) {
					block = loadBlock(con, rs);
				}
				return block;
			}
		} catch (SQLException e) {
			throw new RuntimeException(e.toString(), e);
		} catch (NxtException.ValidationException e) {
			throw new RuntimeException("Block already in database, id = " + blockId + ", does not pass validation!", e);
		}
	}

	static boolean hasBlock(long blockId) {
		try (Connection con = Db.db.getConnection();
				PreparedStatement pstmt = con.prepareStatement("SELECT 1 FROM block WHERE id = ?")) {
			pstmt.setLong(1, blockId);
			try (ResultSet rs = pstmt.executeQuery()) {
				return rs.next();
			}
		} catch (SQLException e) {
			throw new RuntimeException(e.toString(), e);
		}
	}

	static long findBlockIdAtHeight(int height) {
		try (Connection con = Db.db.getConnection();
				PreparedStatement pstmt = con.prepareStatement("SELECT id FROM block WHERE height = ?")) {
			pstmt.setInt(1, height);
			try (ResultSet rs = pstmt.executeQuery()) {
				if (!rs.next()) {
					throw new RuntimeException("Block at height " + height + " not found in database!");
				}
				return rs.getLong("id");
			}
		} catch (SQLException e) {
			throw new RuntimeException(e.toString(), e);
		}
	}

	static BlockImpl findBlockAtHeight(int height) {
		try (Connection con = Db.db.getConnection();
				PreparedStatement pstmt = con.prepareStatement("SELECT * FROM block WHERE height = ?")) {
			pstmt.setInt(1, height);
			try (ResultSet rs = pstmt.executeQuery()) {
				BlockImpl block;
				if (rs.next()) {
					block = loadBlock(con, rs);
				} else {
					throw new RuntimeException("Block at height " + height + " not found in database!");
				}
				return block;
			}
		} catch (SQLException e) {
			throw new RuntimeException(e.toString(), e);
		} catch (NxtException.ValidationException e) {
            throw new RuntimeException("Block already in database at height " + height + ", does not pass validation!", e);
        }
    }

    static BlockImpl findLastBlock() {
        try (Connection con = Db.db.getConnection();
             PreparedStatement pstmt = con.prepareStatement("SELECT * FROM block ORDER BY timestamp DESC LIMIT 1")) {
            BlockImpl block = null;
            try (ResultSet rs = pstmt.executeQuery()) {
                if (rs.next()) {
                    block = loadBlock(con, rs);
                }
            }
            return block;
        } catch (SQLException e) {
            throw new RuntimeException(e.toString(), e);
        } catch (NxtException.ValidationException e) {
            throw new RuntimeException("Last block already in database does not pass validation!", e);
        }
    }

    static BlockImpl findLastBlock(int timestamp) {
        try (Connection con = Db.db.getConnection();
             PreparedStatement pstmt = con.prepareStatement("SELECT * FROM block WHERE timestamp <= ? ORDER BY timestamp DESC LIMIT 1")) {
            pstmt.setInt(1, timestamp);
            BlockImpl block = null;
            try (ResultSet rs = pstmt.executeQuery()) {
                if (rs.next()) {
                    block = loadBlock(con, rs);
                }
            }
            return block;
        } catch (SQLException e) {
            throw new RuntimeException(e.toString(), e);
        } catch (NxtException.ValidationException e) {
            throw new RuntimeException("Block already in database at timestamp " + timestamp + " does not pass validation!", e);
        }
    }

	static BlockImpl loadBlock(Connection con, ResultSet rs) throws NxtException.ValidationException {
		try {
			int version = rs.getInt("version");
			int timestamp = rs.getInt("timestamp");
			long previousBlockId = rs.getLong("previous_block_id");
			long totalAmountNQT = rs.getLong("total_amount");
			long totalFeeNQT = rs.getLong("total_fee");
			int payloadLength = rs.getInt("payload_length");
			byte[] generatorPublicKey = rs.getBytes("generator_public_key");
			byte[] previousBlockHash = rs.getBytes("previous_block_hash");
			BigInteger cumulativeDifficulty = new BigInteger(rs.getBytes("cumulative_difficulty"));
			long baseTarget = rs.getLong("base_target");
			long nextBlockId = rs.getLong("next_block_id");
			int height = rs.getInt("height");
			byte[] generationSignature = rs.getBytes("generation_signature");
			byte[] blockSignature = rs.getBytes("block_signature");
			byte[] payloadHash = rs.getBytes("payload_hash");
			long id = rs.getLong("id");
            return new BlockImpl(version, timestamp, previousBlockId, totalAmountNQT, totalFeeNQT, payloadLength, payloadHash,
                    generatorPublicKey, generationSignature, blockSignature, previousBlockHash,
					cumulativeDifficulty, baseTarget, nextBlockId, height, id);
		} catch (SQLException e) {
			throw new RuntimeException(e.toString(), e);
		}
	}

	static void saveBlock(Connection con, BlockImpl block) {
        try {
            try (PreparedStatement pstmt = con.prepareStatement("INSERT INTO block (id, version, timestamp, previous_block_id, "
                    + "total_amount, total_fee, payload_length, generator_public_key, previous_block_hash, cumulative_difficulty, "
                    + "base_target, height, generation_signature, block_signature, payload_hash, generator_id) "
                    + " VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)")) {
                int i = 0;

                pstmt.setLong(++i, block.getId());
                pstmt.setInt(++i, block.getVersion());
                pstmt.setInt(++i, block.getTimestamp());
                DbUtils.setLongZeroToNull(pstmt, ++i, block.getPreviousBlockId());
                pstmt.setLong(++i, block.getTotalAmountNQT());
                pstmt.setLong(++i, block.getTotalFeeNQT());
                pstmt.setInt(++i, block.getPayloadLength());
                pstmt.setBytes(++i, block.getGeneratorPublicKey());
                pstmt.setBytes(++i, block.getPreviousBlockHash());
                pstmt.setBytes(++i, block.getCumulativeDifficulty().toByteArray());
                pstmt.setLong(++i, block.getBaseTarget());
                pstmt.setInt(++i, block.getHeight());
                pstmt.setBytes(++i, block.getGenerationSignature());
                pstmt.setBytes(++i, block.getBlockSignature());
                pstmt.setBytes(++i, block.getPayloadHash());
                pstmt.setLong(++i, block.getGeneratorId());
                pstmt.executeUpdate();
                TransactionDb.saveTransactions(con, block.getTransactions());
            }
            if (block.getPreviousBlockId() != 0) {
                try (PreparedStatement pstmt = con.prepareStatement("UPDATE block SET next_block_id = ? WHERE id = ?")) {
                    pstmt.setLong(1, block.getId());
                    pstmt.setLong(2, block.getPreviousBlockId());
                    pstmt.executeUpdate();
                }
            }
        } catch (SQLException e) {
            throw new RuntimeException(e.toString(), e);
        }
    }

    // relying on cascade triggers in the database to delete the transactions for all deleted blocks
    static void deleteBlocksFrom(long blockId) {
        if (!Db.db.isInTransaction()) {
            try {
                Db.db.beginTransaction();
                deleteBlocksFrom(blockId);
                Db.db.commitTransaction();
            } catch (Exception e) {
                Db.db.rollbackTransaction();
                throw e;
            } finally {
                Db.db.endTransaction();
            }
            return;
        }
        try (Connection con = Db.db.getConnection();
             PreparedStatement pstmtSelect = con.prepareStatement("SELECT db_id FROM block WHERE timestamp >= "
                     + "(SELECT timestamp FROM block WHERE id = ?) ORDER BY timestamp DESC");
				PreparedStatement pstmtDelete = con.prepareStatement("DELETE FROM block WHERE db_id = ?")) {
			try {
				pstmtSelect.setLong(1, blockId);
				try (ResultSet rs = pstmtSelect.executeQuery()) {
					Db.db.commitTransaction();
					while (rs.next()) {
						pstmtDelete.setLong(1, rs.getLong("db_id"));
						pstmtDelete.executeUpdate();
						Db.db.commitTransaction();
					}
				}
			} catch (SQLException e) {
				Db.db.rollbackTransaction();
				throw e;
			}
		} catch (SQLException e) {
			throw new RuntimeException(e.toString(), e);
		}
	}

	static void deleteAll() {
		if (!Db.db.isInTransaction()) {
			try {
				Db.db.beginTransaction();
				deleteAll();
				Db.db.commitTransaction();
			} catch (Exception e) {
				Db.db.rollbackTransaction();
				throw e;
			} finally {
				Db.db.endTransaction();
			}
			return;
		}
		Logger.logMessage("Deleting blockchain...");
		try (Connection con = Db.db.getConnection();
			 Statement stmt = con.createStatement()) {
			try {
				stmt.executeUpdate("SET REFERENTIAL_INTEGRITY FALSE");
				stmt.executeUpdate("TRUNCATE TABLE transaction");
				stmt.executeUpdate("TRUNCATE TABLE block");
				stmt.executeUpdate("SET REFERENTIAL_INTEGRITY TRUE");
				Db.db.commitTransaction();
			} catch (SQLException e) {
				Db.db.rollbackTransaction();
				throw e;
			}
		} catch (SQLException e) {
			throw new RuntimeException(e.toString(), e);
		}
	}

}
